class NotExtractedException(Exception):
    pass


class EnvelopeAssignedException(Exception):
    pass


class EnvelopeAlreadyMeasuredException(Exception):
    pass


class MissingTemporalProfileArgumentException(Exception):
    pass
